# coding: utf-8

from jinja2 import Environment, PackageLoader

class Jinja2(object):

    def __init__(self, package_name, template_folder='templates'):
        self.loader = PackageLoader(package_name, template_folder)
        self.environment = Environment(loader=self.loader, autoescape=True,
                trim_blocks=True, lstrip_blocks=True)

    def render_template(self, template_name, **data):
        template = self.environment.get_template(template_name)
        return template.render(**data)



# coding: utf-8

import redis
import logging

from reloader.nginx import service_reload_nginx
from reloader.openresty import service_reload_openresty
from reloader.config import config


logging.basicConfig(level=logging.INFO, format='%(levelname)s - %(asctime)s: %(message)s')
logger = logging.getLogger(__name__)

rds = redis.Redis(config.redis_host, config.redis_port)

reloader_function = {
    'nginx': service_reload_nginx,
    'openresty': service_reload_openresty,
}

service_reload = reloader_function[config.reloader_type]


def get_backends(appname):
    entrypoints = rds.hkeys(config.entrypoints_key % appname)
    return {entrypoint: list(rds.smembers(config.backends_key % (appname, entrypoint))) for entrypoint in entrypoints}


def watch():
    pub = rds.pubsub()
    pub.subscribe(config.watch_key)
    try:
        for data in pub.listen():
            if data['type'] != 'message':
                continue
            if data['data'] == 'stop':
                break
            logger.info('Reload Signal of [%s] published, reload' % data['data'])

            appname = data['data']
            backends = get_backends(appname)
            service_reload(appname, backends)
    except Exception as e:
        logger.exception(e)
    finally:
        pub.unsubscribe()

def main():
    try:
        watch()
    except KeyboardInterrupt:
        logger.info('Catch ctrl-c, stopped')



# coding: utf-8

import os
import pwd
import errno

def get_uid(name):
    return pwd.getpwnam(name).pw_uid if isinstance(name, basestring) else name

def get_gid(name):
    return pwd.getpwnam(name).pw_gid if isinstance(name, basestring) else name

def ensure_file(path, owner=None, group=None, mode=0644, content=''):
    try:
        current_content = open(path).read()
    except IOError, e:
        if e.errno == errno.ENOENT:
            current_content = None
        else:
            raise

    if current_content != content:
        with open(path, 'w') as f:
            f.write(content)

    os.chmod(path, mode)
    if owner and group:
        os.chown(path, get_uid(owner), get_gid(group))

def ensure_file_absent(path):
    if os.path.lexists(path):
        os.unlink(path)

def ensure_dir(path, owner=None, group=None, mode=0755):
    try:
        os.mkdir(path, mode)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise

    if owner and group:
        os.chown(path, get_uid(owner), get_gid(group))



# coding: utf-8

import sys
import yaml
from optparse import OptionParser
from tabulate import tabulate


class Config(dict):

    def __getattr__(self, name):
        return self.get(name, None)


_default_config = {
    'watch_key': 'eru:discovery:published',
    'entrypoints_key': 'eru:app:%s:backends',
    'backends_key': 'eru:app:%s:entrypoint:%s:backends',
    'nginx': 'nginx',
    'upstream_dir': '/etc/nginx/up.conf.d/',
    'server_dir': '/etc/nginx/s.conf.d/',
    'redis_host': '127.0.0.1',
    'redis_port': 6379,
    'daemonize': False,
    'pod': 'intra',
    'log_prefix': '/data/logs',
    'template_folder': 'templates',

    'reloader_type': 'nginx',
    'ainur_url': '',
}


def load_config():
    parser = OptionParser()
    parser.add_option('-c', '--config', dest='config_path', type='str', default=None)
    parser.add_option('-w', '--watch-key', dest='watch_key', type='str', default=None)
    parser.add_option('-e', '--entrypoints-key', dest='entrypoints_key', type='str', default=None)
    parser.add_option('-b', '--backends-key', dest='backends_key', type='str', default=None)
    parser.add_option('-n', '--nginx', dest='nginx', type='str', default=None)
    parser.add_option('-u', '--upstream-dir', dest='upstream_dir', type='str', default=None)
    parser.add_option('-s', '--server-dir', dest='server_dir', type='str', default=None)
    parser.add_option('-r', '--redis-host', dest='redis_host', type='str', default=None)
    parser.add_option('-p', '--redis-port', dest='redis_port', type='int', default=None)
    parser.add_option('-d', '--daemon', dest='daemonize', action='store_true', default=None)
    parser.add_option('-o', '--pod', dest='pod', type='str', default='intra')
    parser.add_option('-l', '--log-prefix', dest='log_prefix', type='str', default='/data/logs')
    parser.add_option('-t', '--template-folder', dest='template_folder', type='str', default='templates')
    parser.add_option('-k', '--reloader-type', dest='reloader_type', type='str', default=None)
    parser.add_option('-a', '--ainur-url', dest='ainur_url', type='str', default=None)
    options, _ = parser.parse_args()

    file_config = _default_config
    if options.config_path:
        try:
            with open(options.config_path, 'r') as f:
                file_config.update(yaml.load(f))
        except IOError as e:
            print 'Failed to open %s:\n  %s\nignored\n' % (options.config_path, e)

    for k in _default_config.keys():
        cmd_value = getattr(options, k, '')
        if cmd_value is not None:
            file_config[k] = cmd_value

    if file_config['reloader_type'] not in ('nginx', 'openresty'):
        print 'reloader_type must be in nginx / openresty'
        sys.exit(-1)

    if file_config['reloader_type'] == 'openresty' and not file_config['ainur_url']:
        print 'When use openresty, must set ainur url'
        sys.exit(-1)

    print tabulate([[k, str(v)] for k, v in file_config.iteritems()], headers=['Config Key', 'Config Value'])
    return Config(**file_config)


config = load_config()



# coding: utf-8

import json
import logging
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import RequestException

from reloader.config import config

logger = logging.getLogger(__name__)

_backends_cache = {}
_session = requests.Session()
_session.mount('http://', HTTPAdapter(max_retries=10))


def _backends_no_diff(old, new):
    return set(old) == set(new)


def get_balancers(appname, entrypoint):
    url = '%s/ajax/lb/%s/%s/' % (config.ainur_url, appname, entrypoint)
    if not url.startswith('http://'):
        url = 'http://' + url
    try:
        resp = _session.get(url)
        return resp.json()
    except RequestException as e:
        # TODO 这里其实应该发个邮件/短信之类的通知获取失败
        logger.exception('Error when getting balancer for %s:%s' % (appname, entrypoint))
        logger.exception(e)
        return None


def update_balancer(upstream_url, backend_name, servers):
    if not servers:
        try:
            _session.delete(upstream_url, data=json.dumps({'backend': backend_name}))
        except RequestException as e:
            # TODO 这里也应该通知
            logger.exception('Error when deleting backends for %s' % backend_name)
            logger.exception(e)
        return

    servers = ['server %s;' % b for b in servers]
    try:
        _session.put(upstream_url, data=json.dumps({'backend': backend_name, 'servers': servers}))
    except RequestException as e:
        # TODO 这里也应该通知
        logger.exception('Error when updating backends for %s' % backend_name)
        logger.exception(e)


def service_reload_openresty(appname, backends):
    for entrypoint, servers in backends.iteritems():
        r = get_balancers(appname, entrypoint)
        if not r:
            continue

        backend_name = r['backend_name']
        old_backends = _backends_cache.get(backend_name, [])
        if _backends_no_diff(old_backends, servers):
            logger.info('No new nodes found for [%s], ignore' % appname)
            continue

        _backends_cache[backend_name] = servers

        upstream_urls = [b['lb_client']['upstream_addr'] for b in r['balancers']]
        for upstream_url in upstream_urls:
            update_balancer(upstream_url, backend_name, servers)



# coding: utf-8

import os
import logging
from subprocess import check_call

import reloader
from reloader.config import config
from reloader.ensure import ensure_file, ensure_file_absent, ensure_dir
from reloader.templates import Jinja2


logger = logging.getLogger(__name__)
template = Jinja2(reloader.__name__, template_folder=config.template_folder)

_backends_cache = {}


def _judge_backends_diff(old, new):
    if set(old.keys()) != set(new.keys()):
        return True
    for entrypoint in new.keys():
        old_backends = old[entrypoint]
        new_backends = new[entrypoint]
        if set(old_backends) != set(new_backends):
            return True
    return False


def service_reload_nginx(appname, backends):
    old_backends = _backends_cache.get(appname, {})
    if not _judge_backends_diff(old_backends, backends):
        logger.info('No new nodes found for [%s], ignore' % appname)
        return
        
    _backends_cache[appname] = backends

    # ensure nginx access/error log dir
    ensure_dir(os.path.join(config.log_prefix, appname))

    # reload nginx
    reload_nginx_config(appname, backends)
    reload_nginx()
    
    logger.info('Reload of [%s] done' % appname)


def reload_nginx_config(appname, backends):

    def _check_backends(b):
        """check if backends is not {'entry1': [], 'entry2': []}..."""
        return any(b.values())

    up_conf = os.path.join(config.upstream_dir, '{0}.conf'.format(appname))
    server_conf = os.path.join(config.server_dir, '{0}.conf'.format(appname))
    if not _check_backends(backends):
        ensure_file_absent(up_conf)
        ensure_file_absent(server_conf)
    else:
        upstream_nginx_conf = template.render_template('/upstream_nginx.jinja',
                appname=appname, backends=backends, config=config)
        server_nginx_conf = template.render_template('/server_nginx.jinja',
                appname=appname, backends=backends, config=config)
        ensure_file(up_conf, content=upstream_nginx_conf)
        ensure_file(server_conf, content=server_nginx_conf)


def reload_nginx():
    try:
        check_call([config.nginx, '-s', 'reload'])
    except Exception as e:
        logger.exception('Reload command failed, %s' % e)



# coding: utf-8

from setuptools import setup, find_packages

setup(
    name='eru-reloader',
    version='1.1',
    author='tonic',
    zip_safe=False,
    author_email='tonic@wolege.ca',
    description='A reloader daemon for nginx',
    packages=find_packages(),
    package_data={
        '': ['templates/*.jinja'],
    },
    include_package_data=True,
    entry_points={
        'console_scripts':[
            'rld=reloader.main:main',
            'eru-reloader=reloader.main:main',
        ],
    },
    install_requires=[
        'jinja2',
        'pyyaml',
        'redis',
        'tabulate'
    ],
)
